

; ase2.asm: 12F1822 code, testng part 3, 

; Serial port input and output: 
; 7 - RA0: LED on pin 7, RA0, lit when pin is L, output 0
; 6 - RA1: i2c slave: SCL on RA1 pin 6 
; 5 - RA2: i2c slave: SDA on RA2 pin 5
; 4 - RA3 MCLR
; 3 - RA4: Serial port Transmitter
; 2 - RA5: Serial port Receiver

    RADIX dec
    include p12f1822.inc

; Configuration words

 __CONFIG _CONFIG1, 0x2fe4 
			; 00.1 00 1.1 0 1 0.0 100 	w.dog off 
			; 00.1 00 1.1 0 1 0.1 100 	w.dog controlled by SWDTEN in WDTCON
			; 00.1 00 1.1 1 1 0.0 100 	w.dog controlled by SWDTEN in WDTCON
			; bit 6=1 is mclr enabled
 __CONFIG _CONFIG2, 0x1eff 
			; 0 1.11 10.1111.1111
			; LVP=0
			; bit 8 is PLLEN, 0 means control via OSCCON

; These state variables at 0x70 appear in all 4 banks
 cblock 0x70
    Delay1
    Delay2
    symflags
 endc

; Spaces. Serial reception 16 bytes starting at 0xA0
ser_rd0 EQU 0xA0
ser_rdmask EQU 0xF

 cblock 0x20
    tick61          ; (0x20)
	TMR1C			; (0x21)
    ser_rxfillp     ; (0x22)
    ser_rxreadp     ; (0x23)

	; i2c state stuff
    i2c_command		; (0x24)
    sspstat_Temp	; (0x25)
    i2c_txreadp		; (0x26)
    i2c_rxfillp		; (0x27)
    i2c_rxreadp		; (0x28)
	i2c_eeprom_readp

    seconds610      ; (0x29)
    seconds611      ; (0x2a)
    seconds612      ; (0x2b)
    seconds613      ; (0x2c)

	txhex_hindexkeep
	txhex_hindex
	sectick

 endc


I2C_SLAVEADDR equ 0x29

; i2c reception area. 8 bytes from 0x38 onwards
i2c_rd0 EQU 0x38
i2c_rdmask EQU 0x7

; i2c transmission area. 16 bytes from 0x40 onwards
; Will want 32 bytes eventually...
i2c_td0 EQU 0x40
i2c_tdsize EQU 16

 cblock i2c_td0
    i2c_tx_serial       ; 0x00

    i2c_tx_TMR1L		; 0x01 LSB
    i2c_tx_TMR1H		; 0x02
    i2c_tx_TMR1C		; 0x03 MSB

    ; Time of latest accumulator event
    i2c_tx_seconds610       ; +0x04
    i2c_tx_seconds611       ; +0x05
    i2c_tx_seconds612       ; +0x06
    i2c_tx_seconds613       ; +0x07

    ; Current tick at latest read-action
    i2c_tx_readseconds610   ; +0x08
    i2c_tx_readseconds611   ; +0x09
    i2c_tx_readseconds612   ; +0x0a
    i2c_tx_readseconds613   ; +0x0b

	i2c_tx_eeprombyte	; 0x0c

 endc

; symflags flags:
ONESECOND EQU 0
I2CCWT0 EQU 1

; Reset vector is at 0x0
    ORG 0x0
    goto prog

    ORG 0x4
	goto isrjob

; [
; Beginning of main part.
prog
	banksel INTCON
	clrf INTCON

	; Set oscillator to 8 MHz.
    banksel OSCCON
	movlw 0x70
	movwf OSCCON

	; Osccon register settings for different speeds:
	; 01011000 - 0x58 -  1 MHz
	; 01100000 - 0x60 -  2 MHz
	; 01101000 - 0x68 -  4 MHz
	; 01110000 - 0x70 -  8 MHz
	; 01111000 - 0x78 - 16 MHz
	; 11110000 - 0xf0 - 32 MHz

    ; Prepare the delay counters
    clrf Delay1
    clrf Delay2
    clrf symflags

    ; Turn all ANSELA off, TRISA inputs, WPUAs on
    banksel ANSELA
    clrf ANSELA

    banksel TRISA
	movlw 0x3F
    movwf TRISA
    banksel WPUA
    movwf WPUA

	; No intr-H or intr-L on any of the As for now.
    banksel IOCAN
    clrf IOCAN 
    banksel IOCAP
    clrf IOCAP 

    ; This starts timer0 and timer1
    call timer0_init
    ; call timer1_init

	call serial_init

    call i2c_initvars
    call i2c_init

	; Make a0 a 1-second tick output 
	banksel TRISA
	bcf TRISA,0

    ; Ready to run: Now set GIE
	banksel INTCON
	bsf INTCON,GIE

	; Emit a newline on the serial port... 
    movlw 0x0a 
	call txchar_run

	goto loop

; ]


; [ Interrupt vector is at 0x4. Code begins at 0x5, Just segue into there
isrjob
    ; See what is happening

    ; Start with the i2c receiver -- this seems to be the most 
    ; time-critical item. Then do the frequency counter carry,
    ; the serial port, then the other, possibly slower items.

    banksel PIR1
    btfsc PIR1,SSP1IF    ; we have some i2c interrupt action
    goto service_i2c

    banksel INTCON
    btfsc INTCON,TMR0IF     ; Timer 0 hit
    goto service_timer0    

    banksel PIR1
    btfsc PIR1,TMR1IF
    goto service_timer1   ; Timer1 count overflows

    banksel PIR1
    btfsc PIR1,RCIF       ; we have something on the serial receiver
    goto service_rx 

	; Interrupt-on-changes
	; Pin 3 A4 Accumulator for wind, 1p = 2/3 m
	; Pin 4 A3 Accumulator for Rain, 1p = ??? mm
	;;banksel INTCON
    ;;btfsc INTCON,IOCIF  ; Port line changes
	;;goto service_iocs

	; ADC servicing
    ;;banksel PIR1
    ;;btfsc PIR1, ADIF
    ;;goto service_adc
	
	; Other things could come here
	RETFIE

;[
service_i2c
	; The AN734 has us clearing SSPIF right away
    banksel PIR1
    bcf PIR1,SSP1IF  

    call i2c_irq

	; Release the clock before returning if that hasn't already been done
	banksel SSP1CON1
	bsf SSP1CON1,CKP
	bcf SSP1CON1,SSPOV

	RETFIE
;]

; timer1 overflows carry
service_timer1
    banksel TMR1C
    incf TMR1C,F
    banksel PIR1
    bcf PIR1,TMR1IF
	RETFIE

service_rx  
    call serial_receiver
    banksel PIR1
    bcf PIR1,RCIF		; might actually be done from the act of receiving...
	RETFIE

; Timer 0, 61 Hz frequency, start and stop the frequency-counter
; measurements
service_timer0    
    call one61second_step       ; Count this tick
    banksel tick61
    decfsz tick61,F
    goto donet01
    movlw 61
    movwf tick61

    ; once-a-second action is here
    bsf symflags, ONESECOND

    ; Timer1 is run as a frequency-counter, automatically.
    ; if it isn't on we start it (T1CON,TMR1ON)
    banksel T1CON
    btfss T1CON,TMR1ON
    goto t1isstopped

    bcf T1CON,TMR1ON    ; is set, turn it off and snapshot

    banksel TMR1L
    movf TMR1L,w
    banksel i2c_tx_TMR1L
    movwf i2c_tx_TMR1L
    banksel TMR1H
    movf TMR1H,w
    banksel i2c_tx_TMR1H
    movwf i2c_tx_TMR1H
    banksel TMR1C
    movf TMR1C,w
    banksel i2c_tx_TMR1C
    movwf i2c_tx_TMR1C

    ; Then reset the counters ready for next time
    banksel TMR1H
    clrf TMR1H
    clrf TMR1L
	banksel TMR1C
    clrf TMR1C

    goto donet01

t1isstopped
	banksel T1CON
    bsf T1CON,TMR1ON        ; is stopped, so turn it on

    ; fall into donet01

donet01     
    bcf INTCON,TMR0IF
	RETFIE

; ] End of ISR main functions


; [ Initialize the i2c subsystem

; Set the i2c state variables to sane starting values
i2c_initvars
    banksel i2c_txreadp
    movlw i2c_td0 
    movwf i2c_txreadp
    clrf i2c_rd0
    bcf symflags,I2CCWT0   

    ; i2c Tx buffer: to set on output, the salient values are written starting
    ; at address i2c_td0; i2c_txreadp is set to this address as well before
    ; reading out 

    movlw i2c_rd0
    movwf i2c_rxfillp
    return
; ]

; [ i2c slave init - bus address I2C_SLAVEADDR 
i2c_init
	; Set SDA, SCL to inputs
    banksel TRISA
    bsf TRISA, 1
    bsf TRISA, 2

    ; Make sure the SDA and SCL lines arent occupied by analog inputs
    banksel ANSELA
    bcf ANSELA,1
    bcf ANSELA,2

    ;SSP1CON1   - WCOL SSPOV SSPEN CKP SSPM3 SSPM2 SSPM1 SSPM0
    ;              0    0      1    1    0    1     1      0  <- 7-bit
    banksel SSP1CON1
    movlw 0x36
    movwf SSP1CON1

    ;SSP1CON2   - GCEN ACKSTAT ACKDT ACKEN RCEN PEN RSEN SEN
    ;              0    0      0    0    0    0     0      0 
    ;banksel SSP1CON2
    ;clrf SSP1CON2

    ;SSP1CON3   - ACKTIM PCIE SCIE BOEN SDAHT SBCDE AHEN DHEN 
    ;              0    0      0    0    0    0     0      0 
    ;banksel SSP1CON3
    ;clrf SSP1CON3

    ; SSP1ADD  
    ; slave address times 2 really, so make W contain just that before starting.
    banksel SSP1ADD
    movlw I2C_SLAVEADDR 
    addlw I2C_SLAVEADDR
    movwf SSP1ADD

    ; SSP1STAT  - SMP  CKE   D/A          P    S     R/W          UA        BF
    ;             0    0    data/address Stop Start Read/Write update10 Tx-in-progress
    banksel SSP1STAT
    clrf SSP1STAT

    ; read/reset flag PIR1,SSPIF to see something having come in
    banksel PIR1
    bcf PIR1,SSP1IF

    ; set SSPIE to enable interrupt
    ;   U     ADIE RCIE TXIE SSP1IE CCP1IE TMR2IE TMR1IE 
    ;   0       x    x    x     1     x      x      x 
    banksel PIE1
	bsf PIE1,SSP1IE

    banksel INTCON
	bsf INTCON,PEIE

    ; SSP1BUF     - data in or out
    ; End of i2c preparation
    return
; ] 


; [----- i2c utility functions ------

; Write value from W, avoiding overflows in the process
WriteI2C
    banksel SSP1STAT         ; bank 1
    btfsc SSP1STAT,BF
    goto WriteI2C             ; BF is set so keep waiting
    banksel SSP1CON1          ; bank 0 where SSP1CON1 and SSP1BUF live
DoTheI2CWrite
    bcf SSP1CON1,WCOL         ; Clear WCOL
    movwf SSP1BUF            ; Emit data
    btfsc SSP1CON1,WCOL       ; had collision anyways?
    goto DoTheI2CWrite      ; try again
    bsf SSP1CON1,CKP          ; Set CKP; release the clock
    return  

; Write value from current i2c_txreadp index and increment
WriteCPandI
	; First, we may try to read the sspbuf register so as to clear the BF
    banksel SSP1BUF
    movf SSP1BUF,w

    banksel i2c_txreadp
    movf i2c_txreadp,w
    movwf FSR0L
	clrf FSR0H

	; movlw ':'
	; call txchar_run

    ; movf INDF0,W 
	; call txhex

    ; banksel SSP1STAT    
	; movf SSP1STAT,w
	; call txhex			; See what this might be? 0x0d = 00001101
	   					    ; which are: Start, read, and bf

    movf INDF0,W 
    call WriteI2C		; appears to hang here, predictably... 

	; movlw '.'
	; call txchar_run
    
    ; Increment the read-pointer
    banksel i2c_txreadp
    incf i2c_txreadp,f
    call BoundCurrentP
    return

; Make sure the i2c_txreadp is within bounds, i2c_td0 to i2c_td0+i2c_tdsize-1
BoundCurrentP
    banksel i2c_txreadp

	; less than i2c_td0 ? set to i2c_td0

	movlw i2c_td0			; 
    subwf i2c_txreadp,w		; C=0 when i2c_td0 > i2c_txreadp
    btfss STATUS,C  		; skip if C is set
	goto bound_set_i2c_td0	; do this if C is clear
	
	; greater than i2c_td0+i2c_tdsize-1 ? set to i2c_td0
	movlw i2c_td0+i2c_tdsize 
    subwf i2c_txreadp,w		; C=0 when i2c_td0+i2c_tdsize > i2c_txreadp
    btfss STATUS,C  		; skip if C is set
	return					; C is clear, oke and done

	; fall into this when if C is clear as we're past the edge
bound_set_i2c_td0				
    movlw i2c_td0
    movwf i2c_txreadp
	return

; We have a value in W, set the i2c_txreadp to match 
SetCurrentP
    movwf i2c_txreadp
    call BoundCurrentP
    return

; ] --- end of i2c utility functions ------

; Handle the i2c job proper
; Seems that when talking with the Arduino, the START bit is missing
; in the transaction. Similar things happens with R-pi, though
; there it seems to be a matter of the speed being too great overall. 
; Some discussions about these have been shown, suggesting that 
; this bit is not really needed.  Will be interesting to see what the
; WD-book and the Gumstix will make of this....

; [
i2c_irq
    banksel SSP1STAT
    movf SSP1STAT,w			; W is ssp1stat value

	; Put status into temp before showing next...
    ; banksel sspstat_Temp 
    ; movwf sspstat_Temp
	; movlw 'i'
	; call txchar_run
    ; banksel sspstat_Temp
	; movf sspstat_Temp,w
	; call txhex				; show
    ; banksel sspstat_Temp
	; movf sspstat_Temp,w

	;                     ; 0 0 D/A P S R/W UA BF
	;09       00001001    ; 0 0  A  - S  W  -  BF
	;10       00010001    ; 0 0  A  P -  W  -  - 
	;11       00010001    ; 0 0  A  P -  W  -  BF
	
    ; ignstart: andlw   b'00100101'   ; mask uninteresting bits incl START (bit 3)
    andlw   b'00101101'   ; mask uninteresting bits excl START (bit 3)

    banksel sspstat_Temp        ; Save this for reference
    movwf sspstat_Temp

; 1 Master Write last byte was address: D/A=0 S=1 R/W=0  BF=1 CKP=1 
;         - ..0.10.1 - read SSPBUF to clear BF; this held the device slave address shifted 1 - expect next byte to be the command
checkstate1
    ; ignstart: movlw   b'00000001' 		; Ignoring start bit
    movlw             b'00001001' 		; Not ignoring the start bit
    xorwf sspstat_Temp,w		; x x DA x x RW x BF
    btfss STATUS,Z      		; are we in state 1?
    goto checkstate2    		; no, continue to check for state 2

	; State 1

	; Set the receiver register ready to receive,
	banksel i2c_rxfillp
	movlw i2c_rd0
	movwf i2c_rxfillp

    banksel symflags
    bsf symflags,I2CCWT0      ; set bit 1 in symflags, indicating that we have a command to deal with

    banksel SSP1BUF
    movf SSP1BUF,w       ; read the SSPBUF and discard
    goto done_i2c_isr

; 2 Master Write last byte was data   : D/A=1 S=1 R/W=0  BF=1 CKP=1 
;         - ..1.10.1 - read SSPBUF to clear BF; this is the most recent byte - look at the command expectation flag to see if this is command (if set) or data (if reset). When command, set the read index counter as desired
checkstate2
    ; ignstart: movlw   b'00100001' 		; Ignoring the start bit
    movlw             b'00101001' 		; Not ignoring the start bit
    xorwf sspstat_Temp,w
    btfss STATUS,Z      ; are we in state 2?
    goto checkstate3    ; no, continue to check for state 3

    ; do state2 job

    ; Read SSPBUF and store in reception area
    banksel i2c_rxfillp
    movf i2c_rxfillp,w
    movwf FSR0L
	clrf FSR0H
    
    banksel SSP1BUF
    movf SSP1BUF,w
    movwf INDF0

    ; Increment, roll around after 16 bytes
    banksel i2c_rxfillp
    incf i2c_rxfillp,f
    movlw i2c_rdmask
    andwf i2c_rxfillp,f
    movlw i2c_rd0
    iorwf i2c_rxfillp,f
    goto done_i2c_isr

; 3 Master Read last byte was address : D/A=0 S=1 R/W=1  BF=0 CKP=0 - ..0.11.0 -                                         put current read index value into SSPBUF; then set CKP to release SCL - increment index counter
; 3a Master Read last byte was address: D/A=0 S=1 R/W=1  BF=1 CKP=0 - ..0.11.1 - read SSPBUF to clear BF; discard - then put current read index value into SSPBUF; then set CKP to release SCL - increment index counter
checkstate3 
    movf sspstat_Temp,w
    andlw b'00101100'
    ; ignstart: andlw   b'00100100'	; ignoring START
    xorlw b'00001100' 
    ; ignstart: xorlw   b'00000100' ; ignoring START
    btfss STATUS,Z      ; are we in state 3?
    goto checkstate4    ; no, continue to check for state 4

    ; do state3 job
    call WriteCPandI
    goto done_i2c_isr

; 4 Master Read last byte was data    : D/A=1 S=1 R/W=1  BF=0 CKP=0 - ..1.11.0 -                                         put current read index value into SSPBUF; then set CKP to release SCL - increment index counter
checkstate4
    banksel SSP1CON1
    btfsc SSP1CON1,CKP
    goto checkstate5

    movlw b'00101100'
    ; ignstart: movlw   b'00100100'
    xorwf sspstat_Temp,w
    btfss STATUS,Z      ; are we in state 4?
    goto checkstate5    ; no, continue to check for state 5

    call WriteCPandI    ; state 4 action is same as state 3 action... 
    goto done_i2c_isr

; 5 Master NACK, didn't want anymore  : D/A=1 S=1 R/W=0  BF=0 CKP=1 - ..1.10.0 - The master didn't want any more data that we were sending.  
checkstate5
    movf sspstat_Temp,w
    andlw b'00101000'
    ; ignstart: andlw   b'00100000'
    xorlw b'00101000'
    ; ignstart: xorlw   b'00100000'
    btfss STATUS,Z
    goto error_i2c
    goto done_i2c_isr

; Something else, we can't really do that much? 
; Reset all the flags and bits to sane states and hope for the best.... 
error_i2c 
    banksel SSP1CON1
    bcf SSP1CON1,WCOL         ; Reset the error flags
    bcf SSP1CON1,SSPOV
    bsf SSP1CON1,CKP          ; Set CKP; so as to release the clock
    ; fall into goto done_i2c_isr

done_i2c_isr
    return 

; ]-end of the big i2c-isr function 



; [----- serial port receiver ------
; Serial port receiver action. Grab byte and stuff into *ser_rxfillp
; then increment ser_rxfillp
serial_receiver
    banksel ser_rxfillp
    movf ser_rxfillp,w
    movwf FSR0L
	clrf FSR0H
    
    banksel RCREG
    movf RCREG,W
    movwf INDF0

    ; Put it here as well, gives us something to look at
    banksel i2c_tx_serial
    movwf i2c_tx_serial

    ; Increment, roll around after 16 bytes
    banksel ser_rxfillp
    incf ser_rxfillp,f
    movlw ser_rdmask
    andwf ser_rxfillp,f
    movlw ser_rd0
    iorwf ser_rxfillp,f
    return 

; Wait on tx-ready for serial port transmission
txwait
    banksel TXSTA 
    bsf TXSTA,TXEN      ; Turn transmitter on if it wasn't already on
txcharwreadyck  
    btfss TXSTA,TRMT    ; Wait until this bit is reset, indicating tx empty
    goto txcharwreadyck 
    banksel TXREG
    return

; Transmit character from W but only if we are in non-startup mode

txchar_run
    ; banksel symflags
    ; btfss symflags,STARTUPMODE      ; skip next if clear (startup) set (run)
    ; return                          ; not set, ie clear starutpmode, ie, set, run-mode
    call txwait                     ; set, run-mode so emit the character.
    movwf TXREG
    return

; Wait for and receive a serial-port character
; Return with the character in W
serial_reset
    ; check if OERR is set. If so, stop and restart the RCSTA 
    banksel RCSTA
    btfss RCSTA,OERR
    goto oerrclear  ; skip if OERR was set, need to flick CREN
    bcf RCSTA,CREN ; stop continous reception
    bsf RCSTA,CREN ; restart continous reception, clears OERR 
oerrclear
    banksel RCSTA
    btfss RCSTA,FERR
    goto ferrclear  ; skip if FERR was set, need to re-read RCREG
    ; fix framing error: read RCREG and discard (or have it sitting in w on return...) 
    banksel RCREG
    movf RCREG,w
ferrclear
    return

; Initialize the async serial port
serial_init

	; Set the fill and read pointers for the serial receiver!
    movlw ser_rd0
	banksel ser_rxfillp
	movwf ser_rxfillp
    movwf ser_rxreadp

	; APFCON: Set rx to RA5 and TX to RA4 
	banksel APFCON
	bsf APFCON,RXDTSEL
	bsf APFCON,TXCKSEL

    ; Make sure the TX line isn't set to be analog input on RA4
    banksel ANSELA
    bcf ANSELA,4

    ; Set IO directions
    banksel TRISA
    bsf TRISA, 5		; rx is input
    bcf TRISA, 4		; tx is output

    ; set the startup/run flag to startup
    ; banksel symflags
    ; bcf symflags,STARTUPMODE    

    ; Setup serial transmitter and receiver
    ; CSRC TX9 TXEN SYNC SENDB BRGH TRMT TX9D
    ;  0    0   0-1   0     0    0    0     0
    movlw 0x04			; 0x04 when BRGH is to be set
    banksel TXSTA
    movwf TXSTA

    ; SPEN RX9 SREN CREN ADDEN FERR OERR RX9D 
    ;  0-1   0    0   0-1   0    0    0    0 
    banksel RCSTA
    clrf RCSTA      ; Make sure to toggle SPEN, and CREN

	; Begin baud-rate setting
    ; Set Baud-rate, n=0:SPBRGL, N=SPBRGH:SPBRGL
    ; BRG16 BRGH 
    ;   0    0    BR=Fosc/(64 * (n+1)) -- N=51 Fosc=8 MHz = 2400 Baud
    ;   1    0    BR=Fosc/(16 * (N+1)) -- N=51 Fosc=8 MHz = 9600 Baud

    ;   0    0    BR=Fosc/(64 * (n+1)) 
    ;   0    1    BR=Fosc/(16 * (n+1)) 
    ;   1    0    BR=Fosc/(16 * (N+1))
    ;   1    1    BR=Fosc/(4 * (N+1)) 


    ;   1    0    BR=Fosc/(16 * (N+1)) -- N=25 for 4 MHz clock and 9600 Baud
    ;   1    1    BR=Fosc/(4 * (N+1)) -- N=51 for 8 MHz clock gives 38461 Baud

    banksel SPBRGH
    clrf SPBRGH     ; set SPBRGH to 0 
    movlw 51        
    banksel SPBRGL
    movwf SPBRGL

	; BAUDCON. SCKP=0 for normal polarity
    ; ABDOVF RCIDL 0 SCKP BRG16 0 WUE ABDEN
    ;   0      0   0   0   1    0  0   0
    banksel BAUDCON
    movlw 0x08
    movwf BAUDCON
	; end baud-rate setting

    ; Set SPEN 
    banksel RCSTA
    bsf RCSTA,SPEN

    ; Prepare for rx interrupting (GIE set last)
    ;   U     ADIE RCIE TXIE SSPIE CCP1IE TMR2IE TMR1IE 
    ;   0       x   1    0     x     x      x      x 
    banksel PIE1
    bsf PIE1,RCIE
    bcf PIE1,TXIE
    banksel INTCON
    bsf INTCON,PEIE

    ; clear any receiver-error states: read RCREG and discard, will also start the receiver
    banksel RCREG
    movf RCREG,w

    ; Turn Tx and Rx on
    banksel RCSTA
    bsf RCSTA,CREN
    banksel TXSTA
    bsf TXSTA, TXEN
    return 


; Get the latest serial received data in
serial_latestw
    banksel ser_rxreadp
    movf ser_rxreadp,w
    movwf FSR0L
	clrf FSR0H
    
    incf ser_rxreadp,f
    movlw ser_rdmask
    andwf ser_rxreadp,f
    movlw ser_rd0
    iorwf ser_rxreadp,f

    movf INDF0,W
    return

; Indicate if the receiver has anything. If the empty-pointer is different
; from the fill-pointer we do have some things
; return Z so can check BTFSS STATUS,Z afterwards
rx_hasitems
    banksel ser_rxfillp
    movf ser_rxfillp,w
    subwf ser_rxreadp,w
    return



; ] ----serial port functions -----


; [ Time tick counter, every 1/61 second, 32-bit so we don't have to
; worry about overflows...

one61second_step
    banksel seconds610
    incfsz seconds610,f
    return                  ; seconds610 is nonzero so we are done 
    incfsz seconds611,f
    return                  ; seconds611 is nonzero so we are done 
    incfsz seconds612,f
    return                  ; seconds612 is nonzero so we are done 
    incf seconds613,f
    return                  ; incremented seconds613

one61second_reset
    banksel seconds610
    clrf seconds610
    clrf seconds611
    clrf seconds612
    clrf seconds613
    ; continue doing the snapshot action

; snapshot (done on each rain accumulator event)
one61second_snapshot
    banksel seconds610
    movf seconds610,w
    banksel i2c_tx_seconds610
    movwf i2c_tx_seconds610

    banksel seconds611
    movf seconds611,w
    banksel i2c_tx_seconds611
    movwf i2c_tx_seconds611

    banksel seconds612
    movf seconds612,w
    banksel i2c_tx_seconds612
    movwf i2c_tx_seconds612

    banksel seconds613
    movf seconds613,w
    banksel i2c_tx_seconds613
    movwf i2c_tx_seconds613
    return

; snapshot for reading
one61second_prepread
    banksel seconds610  ; all in 0x20-0x6f area
    movf seconds610,w
    banksel i2c_tx_readseconds610
    movwf i2c_tx_readseconds610
    banksel seconds611
    movf seconds611,w
    banksel i2c_tx_readseconds611
    movwf i2c_tx_readseconds611
    banksel seconds612
    movf seconds612,w
    banksel i2c_tx_readseconds612
    movwf i2c_tx_readseconds612
    banksel seconds613
    movf seconds613,w
    banksel i2c_tx_readseconds613
    movwf i2c_tx_readseconds613
    return

; ] End time tick counter


; [ Indexed access to eeprom: enter with address in w, return
; with the data at that address in w.  Note that this does clobber 
; the bank flags
readindexedeprom
    banksel EEADR
    movwf EEADR

    banksel EECON1
    bsf EECON1,RD
    ; read data from EEPROM
    banksel EEDAT
    movf EEDAT,w
    return
; ]

; [ Utility-functions

; Delay-loops spin for about 257*769 cycles, 197633 cycles
; At Fosc=4MHz (0.25 mys) this will be 49.4 milliseconds
delay64k
    decfsz Delay1,F     ; 1*255 cycles, plus 2 cycles when fall in = 257
    goto delay64k       ; 2*256 cycles = 512  
                        ; The above is repeated 256 times, 256*769

    decfsz Delay2,F     ; 1*255 plus 2 = 257
    goto delay64k       ; 2*256 cycles = 512 + plus 769
    return

delay21                 ; 11*769 cycles = 2.1 ms
    movlw 10
    movwf Delay2
delay21_intl
    decfsz Delay1,F         ; 1*255 cycles, plus 2 cycles when fall in = 257
    goto delay21_intl       ; 2*256 cycles = 512  
                            ; This is repeated 10 times, 10*769

    decfsz Delay2,F         ; 1*255 plus 2 = 257
    goto delay21_intl       ; 2*256 cycles = 512 + plus 769
    return

; Spin 256 times. Will be approx 768 clock cycles. 
; At Fosc=4MHz, (0.25 mys) this is 192.25 microseconds
delay256
    decfsz Delay1,F  ; 1 cycle, or 2 cycles if PC is modified (next instr. is skipped)
    goto delay256    ; 2 cycles
    return

; Delay for about 20 microseconds
delay20
    movlw 25
    movwf Delay1
delay20_intl
    decfsz Delay1,F
    goto delay20_intl
    return

;[

timer0_init
    banksel tick61
    movlw 61
    movwf tick61
    bcf symflags, ONESECOND

    call one61second_reset

    banksel OPTION_REG

    ; RAPBU INTEDG T0CS T0SE PSA PS2 PS1 PS0
    ;  0      1     0     0   0   1   1   0
    movlw 0x46
    ; 8 MHz Sets prescaler to 128, TMR0IF rate is tosc/(4*128*256) = tosc/131072
    ; for a 61.035 rate.

    movwf OPTION_REG

    ; Turn Timer0 interrupts on.
    banksel INTCON
    bsf INTCON,TMR0IE
    return
;]

;[ Prepare timer1 to use as a counter, input on pin 2, T1CKI
timer1_init

    ; Clear the timer1 overflow and registers
    banksel TMR1H
    clrf TMR1L
    clrf TMR1H
    banksel TMR1C
    clrf TMR1C

    banksel T1CON  ; TIGINV TMR1GE T1CKPS1 T1CKPS0 T1OSCEN T1SYNC TMR1CS TMR1ON
    movlw 0x02  ;      0      0        0     0        0      0      1      0
    movwf T1CON

    ; Enable the overflow interrupts
    banksel PIR1
    bcf PIR1,TMR1IF
    banksel PIE1
    bsf PIE1,TMR1IE

    return
;]




; i2c commands defined:

LSERIAL_0 equ 0x12
SERIALOUT_1 equ 0x13

EEPROMREAD_0 EQU 0x1A
EEPROMADDRSET_1 EQU 0x18
EEPROMWRITE_2 EQU 0x19
SPACERD EQU i2c_td0


loop
    btfss symflags,I2CCWT0    ; bit set indicates that there is a command having come in
    goto i2cfinished
    
    ; we have some i2c action
    ; act on different values of i2c_rd0, 

	; Turns out, we should delay here a bit since the i2c action might still
	; be going on pulling in arguments for this command. 
	call delay21 		; wait for 2.1 milliseconds 

    ; Do the commands without parameters first; this gives a little time for those
    ; that have parameters to get them in later on.
    banksel i2c_rd0

    movlw EEPROMREAD_0
    subwf i2c_rd0,w
    btfsc STATUS,Z  
	goto perform_eepromread_0

    ; Save latest serial-port value at i2c_tx_serial
    movlw LSERIAL_0
    banksel i2c_rd0
    subwf i2c_rd0,w
    btfsc STATUS,Z  
    goto perform_lserial

    ; SPACERD 0x40: Space read: 0x40-0x5F, set i2c_txreadp to this address bounded
	; by the low 5 bits
    movf i2c_rd0, w
    andlw b'11100000'           ; AND with 0xE0 leaves high 3 bits
	sublw SPACERD				; SPACERD - w, which should be 0x40
    btfsc STATUS,Z  
    goto perform_spaceread_aim

    ; Now for the commands with 1 argument

    banksel i2c_rd0
    movlw EEPROMADDRSET_1
    subwf i2c_rd0,w
    btfsc STATUS,Z  
	goto perform_eepromaddrset_1

    banksel i2c_rd0
    movlw SERIALOUT_1
    subwf i2c_rd0,w
    btfsc STATUS,Z  
    goto perform_serialout_1

    ; Command with 2 arguments

    banksel i2c_rd0
    movlw EEPROMWRITE_2
    subwf i2c_rd0,w
    btfsc STATUS,Z  
	goto perform_eepromwrite_2

    ; put more commands here 
    ; just emit the command-byte here 

	goto duffcmd

; Indicate command reception failure -- we need to be able to flag
; this, as we see it occasionally for some reason.
duffcmd
	; stuff tbd...
    goto i2cjobdone

; Aim for space-read. If the command is reading it will get this and 
; maybe the next byte
perform_spaceread_aim

	banksel i2c_rd0
    movf i2c_rd0, w		  ; we enter with some address from 0x40-0x5f here
	banksel i2c_txreadp
    movwf i2c_txreadp

    goto i2cjobdone

; get latest serial byte as received
perform_lserial
    call serial_latestw     ; latest serial byte into W
	banksel i2c_tx_serial
    movwf i2c_tx_serial     ; then into the i2c-transmission-space

    movlw i2c_tx_serial 
    banksel i2c_txreadp     ; Aim the read-pointer here too so if we read we get it right away
    movwf i2c_txreadp
    goto i2cjobdone

; Emit a serial byte here
perform_serialout_1           ; emit first byte here 
    banksel i2c_rd0
    movf i2c_rd0+1, w
    call txchar_run
    goto i2cjobdone

; Prepare to read from the eeprom at the address defined, 
; one byte being the eeprom address
perform_eepromaddrset_1
    banksel i2c_rd0
    movf i2c_rd0+1, w
	banksel i2c_eeprom_readp
	movwf i2c_eeprom_readp
    goto i2cjobdone

; read one byte from the eeprom at i2c_eeprom_readp
; put this into i2c_eeprombyte
; increment i2c_eeprom_readp by 1 for each call

perform_eepromread_0
	banksel EECON1
	bcf EECON1,EEPGD
	bcf EECON1,WREN

	; set first address
	banksel i2c_eeprom_readp
	movf i2c_eeprom_readp,w
	banksel EEADR
	movwf EEADR

	banksel EECON1
	bsf EECON1,RD
	; read data from EEPROM
	banksel EEDAT
	movf EEDAT,w
    banksel i2c_tx_eeprombyte
	movwf i2c_tx_eeprombyte

	; next address
	banksel i2c_eeprom_readp
	incf i2c_eeprom_readp,f

    banksel i2c_txreadp
    movlw i2c_tx_eeprombyte
    movwf i2c_txreadp
    goto i2cjobdone

; write to the eeprom at given address
perform_eepromwrite_2
    banksel i2c_rd0  ; low byte is address
    movf i2c_rd0+1, w
	banksel i2c_eeprom_readp
	movwf i2c_eeprom_readp
	banksel EEADR
	movwf EEADR

	banksel i2c_rd0
	movf i2c_rd0+2,w  ; high byte is data
	banksel EEDAT
	movwf EEDAT

	banksel PIR2
	bcf PIR2,EEIF

	; write
	banksel EECON1
	bcf EECON1,EEPGD
	bsf EECON1,WREN

	bcf INTCON,GIE		; disable interrupts
	btfsc INTCON,GIE
	goto $-2

	; magic 0x55 0xaa cycle into EECON2 to make write work
	movlw 0x55
	movwf EECON2
	movlw 0xaa
	movwf EECON2
	; hit the WR command bit ... we're on
	bsf EECON1,WR

	bsf INTCON,GIE		; can turn interrupts back on now
	bcf EECON1,WREN

	banksel PIR2
	; wait until job is done, that is, the flag EEIF in PIR2 is set
	btfss PIR2,EEIF
	goto $-1

	bcf PIR2,EEIF
	; can read to verify? 

    goto i2cjobdone

    ; put more actions in between here

i2cjobdone
    ; Having done the i2c command thing, reset the expectation flag for next turn.
    banksel symflags
    bcf symflags,I2CCWT0      ; clear bit 1 in symflags, indicating that we have a recent command to deal with


i2cfinished

; Serial echo function

 	call rx_hasitems		; Z clear indicates something is there
 	btfsc STATUS,Z
 	goto rxdnone
 	call serial_latestw		; latest into W
 	call txchar_run			; send it out
rxdnone


; Now for the approx once-a-second job
checksecond
	; Check if we have the 1s timeout flag set. 
	btfss symflags, ONESECOND
	goto loop

	; Something here to be done approx once per second.

	; Flip bit A0, read-mod-write LATA 
	banksel LATA
	movf LATA,w
	addlw 0x01
	andlw 0x01
	movwf LATA

	; Squawk hello repeatedly also
	; call sayhello

	; banksel sectick
	; incf sectick,f
	; movf sectick,w
	; call txhex

	; The 1second job is now done.
	bcf symflags, ONESECOND

    ; nothing more happening here 
    goto loop

; [
sayhello
    movlw 'H'
	call txchar_run
    movlw 'e'
	call txchar_run
    movlw 'l'
	call txchar_run
    movlw 'l'
	call txchar_run
    movlw 'o'
	call txchar_run
    movlw '!'
	call txchar_run
	return
; ]


; [ Emit the hex representation of some byte from W to the serial port.
txhex
    banksel txhex_hindexkeep
    movwf txhex_hindexkeep

    swapf txhex_hindexkeep, w
    andlw 0x0F
	banksel txhex_hindex
    movwf txhex_hindex
    call tohexL
    ; emit to serial port   
    call txchar_run

    banksel txhex_hindexkeep
    movf txhex_hindexkeep, w
    andlw 0x0F
	banksel txhex_hindex
    movwf txhex_hindex
    call tohexL
    ; emit to serial port   
    call txchar_run
    banksel txhex_hindexkeep

    return

; convert the value in low 4 bits of variable hindex to a hex-character in W
tohexL
    banksel txhex_hindex
    movlw high hexcodetable     ; Get hex-code list into PCLATH
    movwf PCLATH            
    movlw low hexcodetable      ; Find the address in the hex code table
    addwf txhex_hindex, w

    btfsc STATUS,C          ; If we have rolled over, increment PCLATH
    INCF PCLATH,F           ; as needed.
    movwf PCL               ; Jump into the hexcode table, return from there

hexcodetable
    retlw '0'
    retlw '1'
    retlw '2'
    retlw '3'
    retlw '4'
    retlw '5'
    retlw '6'
    retlw '7'
    retlw '8'
    retlw '9'
    retlw 'a'
    retlw 'b'
    retlw 'c'
    retlw 'd'
    retlw 'e'
    retlw 'f'


; ]


	END





